;
; time_t is seconds UTC since Saturday 2000-01-01 00:00:00
; difference to
;  posix 1970-01-01 946684800
;  ntp   1900-01-01 946684800 + 2208988800
;  gps   1980-01-06 946684800 -  315964800
;
; non-standard range in tm_year: 0..99, and tm_mon: 1..12
;
; XXX mktime_irregular handles only slight irregularity.
;
; XXX year 2100 bug: CENTURY divide not done
; XXX year 2128 bug: tm->year overflow
; XXX year 2400 bug: QCENT divide not done
; XXX "se on jonkun muun ongelma"


	global _gmtime_days
	global _gmtime_r

	global _mktime_regular
	global _mktime_irregular

tm_year equ 0
tm_mon  equ 1
tm_day  equ 2
tm_hour equ 3
tm_min  equ 4
tm_sec  equ 5

	section	.data
	align	1

; days in previous months

_tab_normal
	dc.w 0,31,59,90,120,151,181,212,243,273,304,334,365
_tab_leap
	dc.w 0,31,60,91,121,152,182,213,244,274,305,335,366

QYEAR   equ (366 + 3 * 365)   ; _FIRST_ of those is leap year
CENTURY equ (25 * QYEAR - 1)  ; ... is not
QCENT   equ (4 * CENTURY + 1) ; ... _IS_


	section	.text2 ; XXX not fncallable
	align	1

	; word gmtime_days(time_t *clk, struct tm *result_hms);

_gmtime_days
	xgdy             ; tm ptr

	tste             ; NULL clk ?
	beq  L0
	xgex             ; clk ptr
	lde  0,x
	ldd  2,x         ; not atomic
	bra  L1
L0
	lded time        ; external magic time_t
L1
	lsre             ; 32/16 divide workaround
	rord
	pshm ccr         ; odd second in carry

	ldx  # 86400 / 2
	ediv             ; X is full days, D is offset seconds / 2

	xgex             ; days temporarily to E

	ldx  # 60 / 2    ; D seconds / 2
	idiv
	pulm ccr
	rolb             ; the odd second into proper place
	stab tm_sec,y
	xgdx
	ldx  # 60
	idiv
	stab tm_min,y
	xgdx
	stab tm_hour,y

	xgde
	rts              ; return days

	; void gmtime_r(time_t *clk, struct tm *result);

_gmtime_r

	; pass args as-is, clk* in E, tm* in D
	; tm* returned in IY, days in D

	bsr  _gmtime_days

	; place to get weekday, D % 7 + base_weekday, here

	ldx  # QYEAR
	idiv                  ; full quad years and offset days
	xgex
	asle
	asle                  ; year = days / QYEAR * 4

	ldx  #_tab_leap
	cpd  # 366
	blo  L4               ; skip if in year 0 4 8 ...

	subd # 1              ; shortcut for (offdays - 366) / 365 + 1
	ldx  # 365
	idiv                  ; offset days in year
	aex
	xgex                  ; tm_year += 1 or 2 or 3
	ldx  #_tab_normal
L4
	xgde
	stab tm_year,y
	xgde
	clre                  ; indexing tab[]
L41
	adde # 2
	cpd  e,x
	bhs  L41              ; while days >= tab[]

	aex
	subd -2,x             ; remove previous months from daycount
	addb # 1
	stab tm_day,y

	lsre
	ted
	stab tm_mon,y         ; tm_mon is 1...12

	rts                   ; void return

	; time_t mktime(struct tm *);

	; thanks, lynx mktime. gmtime might benefit from brains too

no_such_year
	clre
	clrd
	rts                ; time_t 0

_mktime_regular
	xgex               ; tm ptr

mktime
	ldab tm_year,x
	bmi  no_such_year  ; 2128 invalid (2100 bombs later)
	clra
	lde  # 365
	emul
	tde                ; tm->year * 365

	clra
	ldab tm_year,x
	lsrb
	lsrb
	ade                ; tm->year / 4
	ldab tm_day,x
	ade                ; days from this month AND ONE EXTRA, see below

	ldd  tm_year,x     ; tm_mon to B, tm_year to A
	ldy  # _tab_normal
	aby
	aby                ; word index
	adde -2,y          ; days from previous months

	anda # 3           ; previous year, leap ?
	bne  L3            ; not 1st year of a quadyear, keep the EXTRA
	cmpb # 2           ; this year, leap ? March or later
	bhi  L3            ; yes, keep the EXTRA day from tm_day

	sube # 1           ; correction, take back the EXTRA day
L3
	ldd  # 86400 / 2
	emul               ; full days to seconds
	asld
	role               ; 16x16 multiply workaround
	pshm e,d           ; need scratch for time_t
	tsy

	lde  # 3600
	ldab tm_hour,x
	clra
	emul
	addd 2,y
	adce 0,y
	std  2,y           ; E not stacked anymore

	ldaa # 60
	ldab tm_min,x
	mul
	addb tm_sec,x
	adca # 0
	addd 2,y
	adce # 0

	ais  #+4
	rts                ; return time_t

	; only month and day are fixed.
	; only reasonable fixes are made.
	; date numbers are handled as signed values.

_mktime_irregular
	xgex
L5
	ldab tm_mon,x
	cmpb # 1
	bge  L6           ; if given month >= 1
	addb # 12
	stab tm_mon,x
	dec  tm_year,x
L6
	cmpb # 12
	ble  L7           ; if given month <= 12
	subb # 12
	stab tm_mon,x
	inc  tm_year,x
L7
	ldaa tm_day,x
	cmpa # 1
	blt  prev_month   ; if given day < 1

	bsr  month_length ; month to daycount, B in/out
	sba
	lble mktime       ; if last day >= given day, month and day now OK.
	staa tm_day,x
	inc  tm_mon,x
	bra  L5           ; tm_day += 31 might skip a full month, redo

prev_month
	decb
	bne  L8
	ldab # 12
	dec  tm_year,x
L8
	stab tm_mon,x
	bsr  month_length ; month to daycount, B in/out
	aba               ; A was negative or 0
	staa tm_day,x
	bra  L5           ; tm_day -= 31 might skip a full month, redo

	; 31,28,31,30,31,30,31,31,30,31,30,31
	;  1  2  3  4  5  6  7  8  9 10 11 12

month_length       ; fails on year 00 when century wraps (year 2100)

	cmpb # 2       ; february ?
	beq  L9
	cmpb # 8       ; carry if month 1...7
	adcb # -1      ; -= 1 if month 8...12
	andb # 1       ; 1 if odd
	addb # 30
	rts
L9
	ldab tm_year,x ; cmpb # (2)100; beq 'ldab #0'
	andb # 3       ; 00 01 02 or 03
	subb # 1       ; carry if year 00 04 08 12 ...
	ldab # 0
	adcb # 28      ; 29 if leap year
	rts

	global _is_last_of_month

_is_last_of_month  ; bool i_l_o_m(struct tm *);
	xgex
	ldab tm_mon,x
	bsr  month_length
	cmpb tm_day,x
	beq  L10       ; non-zero in TRUE case
	clrd           ; or FALSE
L10
	rts

	global _difftime

_difftime          ; long difftime(time_t *a, time_t *b); *a - *b
	tste
	beq  L20
	xgex           ; a
	tstd
	beq  L21
	xgdy           ; b
	lde  0,x
	ldd  2,x
	bra  L22
L20
	xgdy           ; b
	lded time      ; external magic time_t
L22
	subd 2,y
	sbce 0,y
	rts
L21
	lded time      ; external magic time_t
	subd 2,x
	sbce 0,x
	jmp  _lneg


; time_t mktime_gps(dword tow, word week, short leapsec);
; week * 604800 + tow + signed_leapsec

	global _mktime_gps

_mktime_gps

	pshm e,d
	tsx            ;  E,D,PK,PC,week,leapsec

	ldd  8,x       ; week
	lde  # 37800   ; work around 16x16 multiply
	emul
	REPT 4
	asld
	role
	ENDR           ; week * 604800
	addd 2,x
	adce 0,x       ; += tow
	ste  0,x
	std  2,x
	clre           ; spacewise, better than sxtd macro
	ldd  10,x      ; leapsec
	bpl  L30
	come
L30
	addd 2,x
	adce 0,x
	subd # 630720000 & 0FFFFh
	sbce # 630720000 >> 16

	ais  #+4
	rts

