
SCCR0 equ 0FC08h  ; baud rate generator
SCCR1 equ 0FC0Ah
SCCR1_ equ 0FC0Bh

LOOPS equ 4000h   ; 1 = test
WOMS  equ 2000h   ; 1 = TXD pin is OC
ILT   equ 1000h   ; 1 = long idle-line detect
PT    equ 0800h   ; 1 = odd parity, 0 = even
PE    equ 0400h   ; 1 = parity enable
M     equ 0200h   ; 1 = 9 databits, 0 = 8 databits
WAKE  equ 0100h   ; 1 = wake if MSbit set, 0 = wake if line idle
TIE   equ 0080h   ; 1 = TDRE int enable
TCIE  equ 0040h   ; 1 = transmit complete int enable
RIE   equ 0020h   ; 1 = RDRF and OR int enable
ILIE  equ 0010h   ; 1 = idle line int enable
TE    equ 0008h   ; 1 = TXD enable, 0 = disable (TXD is gpio)
RE    equ 0004h   ; 1 = receiver enable
RWU   equ 0002h   ; 1 = wake-up mode (rx data ignored until wake)
SBK   equ 0001h   ; 1 = send break after current frame

SCSR  equ 0FC0Ch
TDRE_ equ 01h     ; transmit data register empty - mask for byte access
TC    equ 80h     ; tranmit complete
RDRF  equ 40h     ; receiver data register full
RAF   equ 20h     ; receiver active, character coming in
IDLE  equ 10h     ; idle line condition detected
OR    equ 08h     ; overrun error
NF    equ 04h     ; noise error
FE    equ 02h     ; framing error
PF    equ 01h     ; parity error

SCDR  equ 0FC0Eh  ; word address

	;--------------------------------------------------------------
	section .bss
	align 1

	BUFLEN = 4096            ; be generous, note rec buf is words XXX
	INSERT = BUFLEN + 0
	REMOVE = BUFLEN + 2
	MASK   = BUFLEN + 4

receive_buffer
	ds.b BUFLEN    ; characters are stored as words XXX 2kB wasted normally
	ds.w 2         ; insert and remove indexes
	ds.w 1         ; character mask

transmit_buffer
	ds.b BUFLEN    ; characters are stored as bytes
	ds.w 2         ; insert and remove indexes


	;--------------------------------------------------------------
	section .inthand
	align 1

SCI_int

	pshm d,e,x,k

SCI_EXIT	MACRO
	pulm d,e,x,k
	rti
	ENDM

	clrb
	tbxk              ; X to ram 0:
	comb
	tbek              ; EXT addressing to F:

	lded SCSR         ; SCDR into D. any orphan OR acked
	ande # RDRF
	beq  L4           ; not rint
	ldx  # receive_buffer
	lde  INSERT,x     ; buffer wraps to empty if not handled timely. drastic
	std  e,x          ; words in buffer
	adde # 2
	ande # BUFLEN - 1
	ste  INSERT,x
	SCI_EXIT          ; rare to get rint+tint at the same time ?
L4
	brclr SCSR, # TDRE_, L5

	ldx  # transmit_buffer
	lde  REMOVE,x
	cpe  INSERT,x
	beq  tx_done
	ldab e,x          ; bytes in buffer
	clra              ; bit 8 is 0
	std  SCDR         ; any command word must be written elsewhere
	adde # 1
	ande # BUFLEN - 1
	ste  REMOVE,x
	SCI_EXIT

tx_done
	bclr SCCR1_, # TIE
L5
	SCI_EXIT

	;--------------------------------------------------------------
	section .text
	align 1

	global _init_SCI
	global _SCI_config, _SCI_getc, _SCI_putc
	global _SCI_canget, _SCI_ssel, _SCI_drain, _SCI_puts

	;--------------------------------------------------------------
_SCI_ssel              ; E = 0 is SERIAL 1, E = other is SERIAL 2

	ldx  # 0           ; wrap-access
	tste
	bne  L9
	bclr 0F907h,x, # 1 ; PORTGP.0
	rts
L9
	bset 0F907h,x, # 1 ; PORTGP.0
	rts

	;--------------------------------------------------------------
_SCI_canget

	ldx  # receive_buffer
	ldd  REMOVE,x
	subd INSERT,x
	rts

	;--------------------------------------------------------------
_SCI_drain

	ldx  # transmit_buffer
L6
	ldd  REMOVE,x
	subd INSERT,x
	bne  L6
	rts

	;--------------------------------------------------------------
_SCI_getc                 ; return -1 or a new character

	ldd  # -1

	ldx  # receive_buffer
	lde  REMOVE,x
	cpe  INSERT,x
	beq  L3
	ldd  e,x
	adde # 2
	ande # BUFLEN - 1
	ste  REMOVE,x
	andd MASK,x
L3
	rts

	;--------------------------------------------------------------
_SCI_putc                ; return 0 if buffer full, nonzero if success
	ted                  ; byte arg

putc_B                   ; must not touch IY
	ldx  # transmit_buffer
	lde  INSERT,x
	stab e,x
	clrd                 ; FAILURE return if D stays at 0
	adde # 1
	ande # BUFLEN - 1
	cpe  REMOVE,x
	beq  L2              ; full, drop
	ste  INSERT,x
	xgdx                 ; IX now 0 and D is goes nonzero
	bset SCCR1_,x, # TIE ; XXX 1st character could be written directly
L2
	rts

_SCI_puts

	xgey                 ; byte *arg
L10
	ldab 0,y
	beq  L11
	aiy  #+1
	bsr  putc_B          ; note IY not pushed between calls
	bra  L10
L11
	rts

	;--------------------------------------------------------------
_init_SCI     ; void init_SCI(word baud_E, enum { none, odd, even } D);
_SCI_config   ; identical function

	pshm ccr

	cpe  # 150           ; something is wrong if baudrate less than 150
	bcc  L0
	lde  # 19200         ; XXX force 19200 baud instead
L0

	ldx  # 0FFh
	ldy  # RIE + TE + RE                ; 0 = 8 none
	cmpb # 0
	beq  L1

	ldy  # RIE + TE + RE + M + PE + PT  ; 1 = 8 odd
	cmpb # 1
	beq  L1

	ldy  # RIE + TE + RE + M + PE       ; 2 = 8 even
	cmpb # 2
	beq  L1

	ldx  # 1FFh
	ldy  # RIE + TE + RE + M            ; 3 = 9 none
L1
	stx  receive_buffer + MASK

	xgex                 ; baud to IX to calc divisor
	lde  # 8
	ldd  # 0
	ediv                 ; 80000h / baud = divisor
	xgex
	adce # 0             ; EDIV leaves carry for easy rounding

	ldd  # SCI_int
	ldx  # 0             ; wrap-access registers

	orp  # 00E0h

	std  00A0h,x         ; vector
	ste  SCCR0,x         ; brg
	sty  SCCR1,x         ; mode etc bits

	pulm ccr
	rts

