
/* single or double precision LLA/ENU reports expected,
 * with MSL altitudes,
 * no configuration is done, strictly rx-only */

#include "standard.h"
#include "fix.h"
#include "time.h"

static
struct flt_scale {
	long units;
	word shift;
}
	float_rad_to_mm = { 27501975, 29 };

static void
shift(dword *p, int shift)
{
	if (shift > -32 && shift < +32) {
		if (shift > 0)
			ASLx(p, shift);
		else
			LSRx(p, -shift);
	} else
		*p = 0; /* zeroes/denormals/infinities/nans all here */
}

static long
tsip_single(byte *f, struct flt_scale *scale)
{
	int  exp = ((*(word *) f) >> 7) & 0xFF;
	long num = ((*(dword *) f) & 0x007FFFFF) | 0x00800000;

	if (scale) {
		exp += scale->shift;
		SCALE32(&num, &scale->units);
	}
	shift(&num, exp - 23 - 127);

	if (f[0] & 0x80)
		lnegate(&num);

	return (num);
}

static long
tsip_double(byte *d, struct flt_scale *scale)
{
	int  exp = ((*(word *) d) >> 4) & 0x7FF;
	long num = ((*(dword *) (d + 1)) & 0x0FFFFFFF) | 0x10000000;

	ASLx(&num, 2);
	num += (d[5] >> 6);

	if (scale) {
		exp += scale->shift;
		SCALE32(&num, &scale->units);
	}
	shift(&num, exp - 30 - 1023);

	if (d[0] & 0x80)
		lnegate(&num);

	return (num);
}

#define INTEGER(i)    (*(word *) &msg[i])
#define SINGLE(i)     tsip_single(&msg[i], NULL)
#define DOUBLE(i)     tsip_double(&msg[i], NULL)

#define SINGLE_COORD(i) tsip_single(&msg[i], &float_rad_to_mm)
#define DOUBLE_COORD(i) tsip_double(&msg[i], &float_rad_to_mm)

static void
tsip_message(byte *msg, byte *end)
{
	static struct vector v;
	static byte valid, fill_1;

	if (debug) {
		print_hexdump(msg, end - msg);
		eol();
	}

	switch (*msg++) {

	case 0x00:                 /* no jumptable, please. (also invalid id) */
		break;

#define HAVE(n) (msg + (n) <= end)

	/* typically once per second, 4A or 84 and then 56 */

	case 0x4A:
		if (HAVE(20)) {
			v.coord.lat = SINGLE_COORD(0);
			v.coord.lon = SINGLE_COORD(4);

			sted_gps_altitude(SINGLE(8));
			valid = 1;
		}
		break;

	case 0x84:
		if (HAVE(36)) {
			v.coord.lat = DOUBLE_COORD(0);
			v.coord.lon = DOUBLE_COORD(8);

			sted_gps_altitude(DOUBLE(16));
			valid = 1;
		}
		break;

	case 0x56:
		if (HAVE(20)) {
			long e = SINGLE(0);
			long n = SINGLE(4);

			v.course = iatan2(e, n);
			v.knots  = muldiv(ihypot(e, n), 3600, 1852);

			if (valid)
				gps_fix(&v, NULL); /* both LLA and ENU valid */
			valid = 0;
		}
		break;

	/* typically once per 5 seconds, after above packets. 6D last */

	case 0x46:                    /* receiver health */
		if (HAVE(2)) {
			if (msg[0]) {         /* 0 = "Doing fixes" */
				valid = 0;
				gps_nofix(NULL);
			}
		}
		break;

	case 0x41:
		if (HAVE(10)) {
			long tow = SINGLE(0);

			if (tow >= 0)
				settime(mktime_gps(tow, INTEGER(4), SINGLE(6)));
		}
		break;

	case 0x6D:
		if (HAVE(17)) {
			if ((msg[0] & 7) == 0)
				valid = 0;

			gps_satellites = (msg[0] >> 4);

			HDOP = SINGLE(5);
			VDOP = SINGLE(9);
		}
		break;
	}
}

void
tsip_gps_input(byte c)
{
	enum { DLE = 0x10, ETX = 0x03, };

	static byte
		nDLE, buf[127],
		*p = buf + 1;             /* until first DLE id, msg[0] is 0x00 */

	if (c == DLE) {
		++nDLE;
		return;                   /* on hold */
	}
	while (nDLE >= 2) {           /* DLE DLE --- a single DLE */
		nDLE -= 2;
		if (p < __ENDOF(buf))
			*p++ = DLE;
	}
	if (nDLE) {                   /* either DLE ETX or DLE id */
		if (c == ETX) {
			tsip_message(buf, p);
			c = 0;                /* until next DLE id, msg[0] is 0x00 */
		}
		nDLE = 0;
		p = buf;
	}
	if (p < __ENDOF(buf))
		*p++ = c;
}

