
#include "param.h"
#include "systm.h"
#include "malloc.h"
#include "stat.h"
#include "ioctl.h"
#include "socket.h"
#include "socketvar.h"
#include "protosw.h"
#include "domain.h"
#include "mbuf.h"
#include "time.h"
#include "kernel.h"

#include "net/if.h"
#include "net/route.h"
#include "net/af.h"

#include "ax25/ax.h"
#include "ax25/ax25.h"
#include "ax25/ax_pcb.h"
#include "ax25/ax_var.h"

u_long	beacon_sendspace = 512;
u_long	beacon_recvspace = 1024;

#define axtoso(ax) ((struct socket *) ((ax)->axp_socket))
#define sotoax(so) ((struct axpcb *) ((so)->so_pcb))

#define soisreadable(so) (sorwakeup(so))
#define sofull(so)       (sbspace(&(so)->so_rcv) <= 0)

beacon_output(axp, m, addr, control)
	struct axpcb *axp;
	struct mbuf *m;
	struct mbuf *addr, *control;
{
	struct ax_addr *dst, *src;
	struct sockaddr_ax *sax;
	struct ifnet *ifp = axp->axp_ifp;
	struct ax_ifaddr *ifa = ax_ifaddr;

	if (control)
		m_freem(control);

	if (addr) {
		sax = mtod(addr, struct sockaddr_ax *);
		if (addr->m_len != sizeof(*sax) ||
			sax->sax_len != sizeof(*sax)) {
			m_freem(m);
			return (EINVAL);
		}
		if (sax->sax_family != AF_AX25) {
			m_freem(m);
			return (EAFNOSUPPORT);
		}
		dst = &sax->sax_addr;
		if (sax->sax_ifname[0]) {
			ifa = ax_ifawithifname(sax->sax_ifname);
			if (ifa)
				ifp = ifa->aa_ifp;
		}
	} else {
		dst = &axp->axp_faddr;
		if (dst->ax_call[0] == 0)
			dst = &ax_broadcastaddr;
	}
	if (axp->axp_laddr.ax_call[0] == 0) {
		if (ifa == NULL) {
			m_freem(m);
			return (EADDRNOTAVAIL);
		}
		src = &ifa->aa_addr.sax_addr;
	}
	if (ifp == NULL) {
		if (ifa == NULL) {
			m_freem(m);
			return (ENETDOWN);
		}
		ifp = ifa->aa_ifp;
	}
	return ax25_send(ifp, dst, src,
		COMMAND, UI, AXTYPE_NOL3, m);
}

beacon_input(ifp, dst, src, digis, ctrl, pid, m)
	struct ifnet *ifp;
	struct ax_addr *dst, *src, *digis;
	u_char ctrl, pid;
	struct mbuf *m;
{
	static struct sockaddr_ax sax = { sizeof(sax), AF_INET, };
	struct axpcb *axp = ax_pcblookupbeacon(ifp, dst);

	if (axp && (ctrl & ~PF) == UI) {
		register struct socket *so = axtoso(axp);

		sax.sax_addr = *src;
		if (sbappendaddr(&so->so_rcv, &sax, m, NULL)) {
			soisreadable(so);
			return;
		}
	}
	m_freem(m);
}

/*ARGSUSED*/
beacon_usrreq(so, req, m, nam, control)
	struct socket *so;
	int req;
	struct mbuf *m, *nam, *control;
{
	register struct axpcb *axp;
	int s, error = 0;

	if (req == PRU_CONTROL)
		return (ax_control(so, (int)m, (caddr_t)nam, (struct ifnet *)control));

	if (control && control->m_len) {
		m_freem(control);
		if (m)
			m_freem(m);
		return (EINVAL);
	}

	s = splnet();
	axp = sotoax(so);

	if (axp == NULL && req != PRU_ATTACH) {
		splx(s);
		return (EINVAL);		/* XXX */
	}
	switch (req) {

	case PRU_ATTACH:
		if (axp) {
			error = EISCONN;
			break;
		}
		error = soreserve(so, beacon_sendspace, beacon_recvspace);
		if (error)
			return (error);
		error = ax_pcballoc(so, &axbeacons);
		if (error)
			break;
		break;

	case PRU_DETACH:
		soisdisconnected(so);
		ax_pcbdetach(sotoax(so));
		break;

	case PRU_BIND:
		error = ax_pcbbind(axp, nam);
		break;

	case PRU_LISTEN:
		error = EOPNOTSUPP;
		break;

	case PRU_CONNECT:
		error = ax_pcbconnect(axp, nam);
		if (error)
			break;
		soisconnected(so);
		break;

	case PRU_CONNECT2:
		error = EOPNOTSUPP;
		break;

	case PRU_DISCONNECT:
		if (axp->axp_ifp == NULL) {
			error = ENOTCONN;
			break;
		}
		axp->axp_ifp = NULL;
		so->so_state &= ~SS_ISCONNECTED;

		error = ax_pcbdisconnect(axp);
		break;

	case PRU_ACCEPT:
		error = EOPNOTSUPP;
		break;

	case PRU_SHUTDOWN:
		socantsendmore(so);
		break;

	case PRU_SEND:
		return (beacon_output(axp, m, nam, control));
		break;

	case PRU_ABORT:
		soisdisconnected(so);
		ax_pcbdetach(sotoax(so));
		break;

	case PRU_SENSE:
		break;

	case PRU_RCVOOB:
		error = EINVAL;
		break;

	case PRU_SENDOOB:
		error = EOPNOTSUPP;
		break;

	case PRU_SOCKADDR:
		ax_setsockaddr(axp, nam);
		break;

	case PRU_PEERADDR:
		ax_setpeeraddr(axp, nam);
		break;

	default:
		panic("beacon_usrreq: unknown request");
	}
	splx(s);
	return (error);
}

beacon_ctloutput(op, so, level, optname, mp)
	int op;
	struct socket *so;
	int level, optname;
	struct mbuf **mp;
{
	int error = 0;
	struct axpcb *axp = sotoax(so);
	register struct mbuf *m;

#ifdef notyet
	if (level != ourproto)
		return (lower_ctloutput(op, so, level, optname, mp));
#endif

	switch (op) {

	case PRCO_SETOPT:
		m = *mp;
		switch (optname) {

		default:
			error = EINVAL;
			break;
		}
		if (m)
			(void) m_free(m);
		break;

	case PRCO_GETOPT:
		*mp = m = m_get(M_WAIT, MT_SOOPTS);
		m->m_len = sizeof(int);

		switch (optname) {
		default:
			error = EINVAL;
			break;
		}
		break;
	}
	return (error);
}

